package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.blacklistedJwtTokens
import dev.fyloz.colorrecipesexplorer.config.defaultGroupCookieName
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.*
import dev.fyloz.colorrecipesexplorer.model.validation.or
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import dev.fyloz.colorrecipesexplorer.repository.GroupRepository
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.security.core.userdetails.UserDetails
import org.springframework.security.core.userdetails.UserDetailsService
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.transaction.Transactional
import org.springframework.security.core.userdetails.User as SpringUser

interface UserService :
    ExternalModelService<User, UserSaveDto, UserUpdateDto, UserOutputDto, UserRepository> {
    /** Check if an [User] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean

    /** Gets the user with the given [id]. */
    fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User

    /** Gets all users which have the given [group]. */
    fun getByGroup(group: Group): Collection<User>

    /** Gets the default user of the given [group]. */
    fun getDefaultGroupUser(group: Group): User

    /** Save a default group user for the given [group]. */
    fun saveDefaultGroupUser(group: Group)

    /** Updates de given [entity]. **/
    fun update(entity: User, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User

    /** Updates the last login time of the user with the given [userId]. */
    fun updateLastLoginTime(userId: Long, time: LocalDateTime = LocalDateTime.now()): User

    /** Updates the password of the user with the given [id]. */
    fun updatePassword(id: Long, password: String): User

    /** Adds the given [permission] to the user with the given [userId]. */
    fun addPermission(userId: Long, permission: Permission): User

    /** Removes the given [permission] from the user with the given [userId]. */
    fun removePermission(userId: Long, permission: Permission): User

    /** Logout an user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest)
}

interface GroupService :
    ExternalNamedModelService<Group, GroupSaveDto, GroupUpdateDto, GroupOutputDto, GroupRepository> {
    /** Gets all the users of the group with the given [id]. */
    fun getUsersForGroup(id: Long): Collection<User>

    /** Gets the default group from a cookie in the given HTTP [request]. */
    fun getRequestDefaultGroup(request: HttpServletRequest): Group

    /** Sets the default group cookie for the given HTTP [response]. */
    fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse)
}

interface CreUserDetailsService : UserDetailsService {
    /** Loads an [User] for the given [id]. */
    fun loadUserById(id: Long, ignoreDefaultGroupUsers: Boolean = false): UserDetails
}

@Service
@Profile("!emergency")
class UserServiceImpl(
    userRepository: UserRepository,
    @Lazy val groupService: GroupService,
    @Lazy val passwordEncoder: PasswordEncoder,
) : AbstractExternalModelService<User, UserSaveDto, UserUpdateDto, UserOutputDto, UserRepository>(
    userRepository
),
    UserService {
    override fun idNotFoundException(id: Long) = userIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = userIdAlreadyExistsException(id)

    override fun User.toOutput() = UserOutputDto(
        this.id,
        this.firstName,
        this.lastName,
        this.group,
        this.flatPermissions,
        this.permissions,
        this.lastLoginTime
    )

    override fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean =
        repository.existsByFirstNameAndLastName(firstName, lastName)

    override fun getAll(): Collection<User> =
        super.getAll().filter { !it.isSystemUser && !it.isDefaultGroupUser }

    override fun getById(id: Long): User =
        getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User =
        super.getById(id).apply {
            if (ignoreSystemUsers && isSystemUser || ignoreDefaultGroupUsers && isDefaultGroupUser)
                throw idNotFoundException(id)
        }

    override fun getByGroup(group: Group): Collection<User> =
        repository.findAllByGroup(group).filter {
            !it.isSystemUser && !it.isDefaultGroupUser
        }

    override fun getDefaultGroupUser(group: Group): User =
        repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)

    override fun save(entity: UserSaveDto): User =
        save(with(entity) {
            User(
                id,
                firstName,
                lastName,
                passwordEncoder.encode(password),
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = if (groupId != null) groupService.getById(groupId) else null,
                permissions = permissions
            )
        })

    override fun save(entity: User): User {
        if (existsById(entity.id))
            throw userIdAlreadyExistsException(entity.id)
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw userFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        return super<AbstractExternalModelService>.save(entity)
    }

    override fun saveDefaultGroupUser(group: Group) {
        save(
            user(
                id = 1000000L + group.id!!,
                firstName = group.name,
                lastName = "User",
                password = passwordEncoder.encode(group.name),
                group = group,
                isDefaultGroupUser = true
            )
        )
    }

    override fun updateLastLoginTime(userId: Long, time: LocalDateTime): User {
        val user = getById(userId, ignoreDefaultGroupUsers = true, ignoreSystemUsers = false)
        user.lastLoginTime = time
        return update(
            user,
            ignoreDefaultGroupUsers = true,
            ignoreSystemUsers = false
        )
    }

    override fun update(entity: UserUpdateDto): User {
        val persistedUser by lazy { getById(entity.id) }
        return update(with(entity) {
            User(
                id = id,
                firstName = firstName or persistedUser.firstName,
                lastName = lastName or persistedUser.lastName,
                password = persistedUser.password,
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = if (entity.groupId != null) groupService.getById(entity.groupId) else persistedUser.group,
                permissions = permissions?.toMutableSet() ?: persistedUser.permissions,
                lastLoginTime = persistedUser.lastLoginTime
            )
        })
    }

    override fun update(entity: User): User =
        update(entity, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun update(entity: User, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User {
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw userFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        }

        return super.update(entity)
    }

    override fun updatePassword(id: Long, password: String): User {
        val persistedUser = getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
        return super.update(with(persistedUser) {
            User(
                id,
                firstName,
                lastName,
                passwordEncoder.encode(password),
                isDefaultGroupUser,
                isSystemUser,
                group,
                permissions,
                lastLoginTime
            )
        })
    }

    override fun addPermission(userId: Long, permission: Permission): User =
        super.update(getById(userId).apply { permissions += permission })

    override fun removePermission(userId: Long, permission: Permission): User =
        super.update(getById(userId).apply { permissions -= permission })

    override fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, "Authorization")
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }
}

const val defaultGroupCookieMaxAge = 10 * 365 * 24 * 60 * 60 // 10 ans

@Service
@Profile("!emergency")
class GroupServiceImpl(
    private val userService: UserService,
    groupRepository: GroupRepository
) : AbstractExternalNamedModelService<Group, GroupSaveDto, GroupUpdateDto, GroupOutputDto, GroupRepository>(
    groupRepository
),
    GroupService {
    override fun idNotFoundException(id: Long) = groupIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = groupIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = groupNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = groupNameAlreadyExistsException(name)

    override fun Group.toOutput() = GroupOutputDto(
        this.id!!,
        this.name,
        this.permissions,
        this.flatPermissions
    )

    override fun existsByName(name: String): Boolean = repository.existsByName(name)
    override fun getUsersForGroup(id: Long): Collection<User> =
        userService.getByGroup(getById(id))

    @Transactional
    override fun save(entity: Group): Group {
        return super<AbstractExternalNamedModelService>.save(entity).apply {
            userService.saveDefaultGroupUser(this)
        }
    }

    override fun update(entity: GroupUpdateDto): Group {
        val persistedGroup by lazy { getById(entity.id) }
        return update(with(entity) {
            Group(
                entity.id,
                if (name.isNotBlank()) entity.name else persistedGroup.name,
                if (permissions.isNotEmpty()) entity.permissions else persistedGroup.permissions
            )
        })
    }

    @Transactional
    override fun delete(entity: Group) {
        userService.delete(userService.getDefaultGroupUser(entity))
        super.delete(entity)
    }

    override fun getRequestDefaultGroup(request: HttpServletRequest): Group {
        val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
            ?: throw NoDefaultGroupException()
        val defaultGroupUser = userService.getById(
            defaultGroupCookie.value.toLong(),
            ignoreDefaultGroupUsers = false,
            ignoreSystemUsers = true
        )
        return defaultGroupUser.group!!
    }

    override fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse) {
        val group = getById(groupId)
        val defaultGroupUser = userService.getDefaultGroupUser(group)
        response.addHeader(
            "Set-Cookie",
            "$defaultGroupCookieName=${defaultGroupUser.id}; Max-Age=${defaultGroupCookieMaxAge}; Path=/api; HttpOnly; Secure; SameSite=strict"
        )
    }
}

@Service
@Profile("!emergency")
class CreUserDetailsServiceImpl(
    private val userService: UserService
) :
    CreUserDetailsService {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserById(username.toLong(), true)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    override fun loadUserById(id: Long, ignoreDefaultGroupUsers: Boolean): UserDetails {
        val user = userService.getById(
            id,
            ignoreDefaultGroupUsers = ignoreDefaultGroupUsers,
            ignoreSystemUsers = false
        )
        return SpringUser(user.id.toString(), user.password, user.authorities)
    }
}
