package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.DatabaseUpdaterProperties
import dev.fyloz.colorrecipesexplorer.SUPPORTED_DATABASE_VERSION
import dev.fyloz.colorrecipesexplorer.config.FileConfiguration
import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.emergencyMode
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.ConfigurationRepository
import org.springframework.boot.info.BuildProperties
import org.springframework.context.annotation.Lazy
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Service
import java.time.LocalDate
import java.time.Period
import java.time.ZoneId
import javax.annotation.PostConstruct

interface ConfigurationService {
    /** Gets all set configurations. */
    fun getAll(): List<Configuration>

    /**
     * Gets all configurations with keys contained in the given [formattedKeyList].
     * The [formattedKeyList] contains wanted configuration keys separated by a semi-colon.
     */
    fun getAll(formattedKeyList: String): List<Configuration>

    /**
     * Gets the configuration with the given [key].
     * If the [key] does not exists, an [InvalidConfigurationKeyException] will be thrown.
     */
    fun get(key: String): Configuration

    /** Gets the configuration with the given [type]. */
    fun get(type: ConfigurationType): Configuration

    /** Sets the content of each configuration in the given [configurations] list. */
    fun set(configurations: List<ConfigurationDto>)

    /**
     * Sets the content of the configuration matching the given [configuration].
     * If the given key does not exists, an [InvalidConfigurationKeyException] will be thrown.
     */
    fun set(configuration: ConfigurationDto)

    /** Sets the content of the configuration with the given [type]. */
    fun set(type: ConfigurationType, content: String)

    /** Sets the content of the configuration matching the given [configuration] with a given image. */
    fun set(configuration: ConfigurationImageDto)
}

const val CONFIGURATION_LOGO_FILE_PATH = "images/logo"
const val CONFIGURATION_ICON_FILE_PATH = "images/icon"
const val CONFIGURATION_FORMATTED_LIST_DELIMITER = ';'

@Service
class ConfigurationServiceImpl(
    @Lazy private val repository: ConfigurationRepository,
    private val fileService: FileService,
    private val fileConfiguration: FileConfiguration,
    private val creProperties: CreProperties,
    private val databaseProperties: DatabaseUpdaterProperties,
    private val buildInfo: BuildProperties
) : ConfigurationService {
    override fun getAll() =
        ConfigurationType.values().mapNotNull {
            try {
                get(it)
            } catch (_: ConfigurationNotSetException) {
                null
            }
        }

    override fun getAll(formattedKeyList: String) =
        formattedKeyList.split(CONFIGURATION_FORMATTED_LIST_DELIMITER).map(this::get)

    override fun get(key: String) =
        get(key.toConfigurationType())

    override fun get(type: ConfigurationType) = when {
        type.computed -> getComputedConfiguration(type)
        type.file -> fileConfiguration.get(type)
        !emergencyMode -> repository.findByIdOrNull(type.key)?.toConfiguration()
        else -> null
    } ?: throw ConfigurationNotSetException(type)

    override fun set(configurations: List<ConfigurationDto>) {
        configurations.forEach(this::set)
    }

    override fun set(configuration: ConfigurationDto) = with(configuration) {
        set(key.toConfigurationType(), content)
    }

    override fun set(type: ConfigurationType, content: String) {
        when {
            type.computed -> throw CannotSetComputedConfigurationException(type)
            type.file -> fileConfiguration.set(type, content)
            !emergencyMode -> repository.save(configuration(type, content).toEntity())
        }
    }

    override fun set(configuration: ConfigurationImageDto) {
        val filePath = when (val configurationType = configuration.key.toConfigurationType()) {
            ConfigurationType.INSTANCE_LOGO_PATH -> CONFIGURATION_LOGO_FILE_PATH
            ConfigurationType.INSTANCE_ICON_PATH -> CONFIGURATION_ICON_FILE_PATH
            else -> throw InvalidImageConfigurationException(configurationType)
        }

        fileService.write(configuration.image, filePath, true)
    }

    @PostConstruct
    fun initializeProperties() {
        ConfigurationType.values().filter { !it.computed }.forEach {
            try {
                get(it)
            } catch (_: ConfigurationNotSetException) {
                set(it, it.defaultContent)
            }
        }
    }

    private val ConfigurationType.defaultContent: String
        get() = when (this) {
            ConfigurationType.INSTANCE_NAME -> "Color Recipes Explorer"
            ConfigurationType.INSTANCE_LOGO_PATH -> "images/logo"
            ConfigurationType.INSTANCE_ICON_PATH -> "images/icon"
            ConfigurationType.INSTANCE_URL -> creProperties.deploymentUrl
            ConfigurationType.DATABASE_URL -> databaseProperties.url
            ConfigurationType.DATABASE_USER -> databaseProperties.username
            ConfigurationType.DATABASE_PASSWORD -> databaseProperties.password
            ConfigurationType.RECIPE_APPROBATION_EXPIRATION -> period(months = 4)
            ConfigurationType.TOUCH_UP_KIT_CACHE_PDF -> "true"
            ConfigurationType.TOUCH_UP_KIT_EXPIRATION -> period(months = 1)
            else -> ""
        }

    private fun getComputedConfiguration(key: ConfigurationType) = configuration(
        key, when (key) {
            ConfigurationType.EMERGENCY_MODE_ENABLED -> emergencyMode
            ConfigurationType.BUILD_VERSION -> buildInfo.version
            ConfigurationType.BUILD_TIME -> LocalDate.ofInstant(buildInfo.time, ZoneId.systemDefault()).toString()
            ConfigurationType.DATABASE_SUPPORTED_VERSION -> SUPPORTED_DATABASE_VERSION
            ConfigurationType.JAVA_VERSION -> Runtime.version()
            ConfigurationType.OPERATING_SYSTEM -> "${System.getProperty("os.name")} ${System.getProperty("os.version")} ${
                System.getProperty(
                    "os.arch"
                )
            }"
            else -> throw IllegalArgumentException("Cannot get the value of the configuration with the key ${key.key} because it is not a computed configuration")
        }.toString()
    )

    private fun period(days: Int = 0, months: Int = 0, years: Int = 0) =
        Period.of(days, months, years).toString()
}
