package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import java.io.File
import java.io.IOException
import java.nio.file.Path
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

private val creProperties = CreProperties().apply {
    dataDirectory = "data"
    deploymentUrl = "http://localhost"
}
private const val mockFilePath = "existingFile"
private val mockFilePathPath = Path.of(mockFilePath)
private val mockFileData = byteArrayOf(0x1, 0x8, 0xa, 0xf)

private class FileServiceTestContext {
    val fileService = spyk(FileServiceImpl(creProperties, mockk {
        every { error(any(), any<Exception>()) } just Runs
    }))
    val mockFile = mockk<File> {
        every { path } returns mockFilePath
        every { exists() } returns true
        every { isFile } returns true
        every { toPath() } returns mockFilePathPath
    }
    val mockFileFullPath = spyk(FilePath("${creProperties.dataDirectory}/$mockFilePath")) {
        every { file } returns mockFile

        with(fileService) {
            every { mockFilePath.fullPath() } returns this@spyk
        }
    }
    val mockMultipartFile = spyk(MockMultipartFile(mockFilePath, mockFileData))
}

class FileServiceTest {
    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    // exists()

    @Test
    fun `exists() returns true when the file at the given path exists and is a file`() {
        test {
            assertTrue { fileService.exists(mockFilePath) }
        }
    }

    @Test
    fun `exists() returns false when the file at the given path does not exist`() {
        test {
            every { mockFile.exists() } returns false

            assertFalse { fileService.exists(mockFilePath) }
        }
    }

    @Test
    fun `exists() returns false when the file at the given path is not a file`() {
        test {
            every { mockFile.isFile } returns false

            assertFalse { fileService.exists(mockFilePath) }
        }
    }

    // read()

    @Test
    fun `read() returns a valid ByteArrayResource`() {
        test {
            whenMockFilePathExists {
                mockkStatic(File::readBytes)
                every { mockFile.readBytes() } returns mockFileData

                val redResource = fileService.read(mockFilePath)

                assertEquals(mockFileData, redResource.byteArray)
            }
        }
    }

    @Test
    fun `read() throws FileNotFoundException when no file exists at the given path`() {
        test {
            whenMockFilePathExists(false) {
                with(assertThrows<FileNotFoundException> { fileService.read(mockFilePath) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    @Test
    fun `read() throws FileReadException when an IOException is thrown`() {
        test {
            whenMockFilePathExists {
                mockkStatic(File::readBytes)
                every { mockFile.readBytes() } throws IOException()

                with(assertThrows<FileReadException> { fileService.read(mockFilePath) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    // create()

    @Test
    fun `create() creates a file at the given path`() {
        test {
            whenMockFilePathExists(false) {
                mockkStatic(File::create)
                every { mockFile.create() } just Runs

                fileService.create(mockFilePath)

                verify {
                    mockFile.create()
                }
            }
        }
    }

    @Test
    fun `create() does nothing when a file already exists at the given path`() {
        test {
            whenMockFilePathExists {
                fileService.create(mockFilePath)

                verify(exactly = 0) {
                    mockFile.create()
                }
            }
        }
    }

    @Test
    fun `create() throws FileCreateException when the file creation throws an IOException`() {
        test {
            whenMockFilePathExists(false) {
                mockkStatic(File::create)
                every { mockFile.create() } throws IOException()

                with(assertThrows<FileCreateException> { fileService.create(mockFilePath) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    // write()

    @Test
    fun `write() creates and writes the given MultipartFile to the file at the given path`() {
        test {
            whenMockFilePathExists(false) {
                every { fileService.create(mockFilePath) } just Runs
                every { mockMultipartFile.transferTo(mockFilePathPath) } just Runs

                fileService.write(mockMultipartFile, mockFilePath, false)

                verify {
                    fileService.create(mockFilePath)
                    mockMultipartFile.transferTo(mockFilePathPath)
                }
            }
        }
    }

    @Test
    fun `write() throws FileExistsException when a file at the given path already exists and overwrite is disabled`() {
        test {
            whenMockFilePathExists {
                with(assertThrows<FileExistsException> { fileService.write(mockMultipartFile, mockFilePath, false) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    @Test
    fun `write() writes the given MultipartFile to an existing file when overwrite is enabled`() {
        test {
            whenMockFilePathExists {
                every { mockMultipartFile.transferTo(mockFilePathPath) } just Runs

                fileService.write(mockMultipartFile, mockFilePath, true)

                verify {
                    mockMultipartFile.transferTo(mockFilePathPath)
                }
            }
        }
    }

    @Test
    fun `write() throws FileWriteException when writing the given file throws an IOException`() {
        test {
            whenMockFilePathExists(false) {
                every { fileService.create(mockFilePath) } just Runs
                every { mockMultipartFile.transferTo(mockFilePathPath) } throws IOException()

                with(assertThrows<FileWriteException> {
                    fileService.write(mockMultipartFile, mockFilePath, false)
                }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    // delete()

    @Test
    fun `delete() deletes the file at the given path`() {
        test {
            whenMockFilePathExists {
                every { mockFile.delete() } returns true

                fileService.delete(mockFilePath)
            }
        }
    }

    @Test
    fun `delete() throws FileNotFoundException when no file exists at the given path`() {
        test {
            whenMockFilePathExists(false) {
                with(assertThrows<FileNotFoundException> { fileService.delete(mockFilePath) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    @Test
    fun `delete() throws FileDeleteException when deleting throw and IOException`() {
        test {
            whenMockFilePathExists {
                every { mockFile.delete() } throws IOException()

                with(assertThrows<FileDeleteException> { fileService.delete(mockFilePath) }) {
                    assertEquals(mockFilePath, this.path)
                }
            }
        }
    }

    // String.fullPath()

    @Test
    fun `fullPath() appends the given path to the given working directory`() {
        test {
            with(fileService) {
                val fullFilePath = mockFilePath.fullPath()

                assertEquals("${creProperties.dataDirectory}/$mockFilePath", fullFilePath.path)
            }
        }
    }

    @Test
    fun `fullPath() throws InvalidFilePathException when the given path contains invalid fragments`() {
        test {
            with(fileService) {
                BANNED_FILE_PATH_SHARDS.forEach {
                    val maliciousPath = "$it/$mockFilePath"

                    with(assertThrows<InvalidFilePathException> { maliciousPath.fullPath() }) {
                        assertEquals(maliciousPath, this.path)
                        assertEquals(it, this.fragment)
                    }
                }
            }
        }
    }

    private fun test(test: FileServiceTestContext.() -> Unit) {
        FileServiceTestContext().test()
    }

    private fun FileServiceTestContext.whenMockFilePathExists(exists: Boolean = true, test: () -> Unit) {
        every { fileService.exists(mockFilePath) } returns exists
        test()
    }
}
